;
;GPIO3 source for MoBoStamp-pe AVR firmware.
;
;(c) 2006 Bueno systems, Inc.
;
;This source code is provided under the TERMS OF USE shown at the bottom of this file
;in the hopes that it will inspire others to contribute back to the library of MoBoStamp-pe
;AVR code. NO SUPPORT OF ANY KIND IS OFFERED, AND NONE WILL BE PROVIDED, beyond that contained
;in the end-user documentation downloadable from the Parallax website.
;
	.NOLIST
	.DEVICE	ATTiny13
	.INCLUDE	"TN13DEF.INC"
	.INCLUDE "\jobs\avr\BSIMACROS.INC"
	.LIST
	.LISTMAC

	.EQU	OWIO    = PINB2

	.DEF	MACREG  = R16
	.DEF	COMMAND = R17
	.DEF	ACC     = R18
	.DEF	PTIME   = R19
	.DEF	BITCNT  = R20
	.DEF	PINMASK = R21

	.DEF	RESULTL = R24
	.DEF	RESULTH = R25
	.DEF	CNTMODE = R26

	.EQU	PINNOH  = 5
	.EQU	PINNOL  = 4
	.EQU	ANALOG  = 3
	.EQU	OUTPUT  = 2
	.EQU	INTREF  = 2
	.EQU	PWM     = 1
	.EQU	USEHI   = 1
	.EQU	USELO   = 0
	.EQU	SETVAL  = 0
	.EQU	INVPWM  = 0

	.EQU	DEVICE0 = 'G'
	.EQU	DEVICE1 = 'P'
	.EQU	VERSION = 3

	.CSEG

;--------Hardware reset.---------------------

	.ORG	0
	RJMP	START

;--------OWIO PC interrupt for counting.-----
	
	.ORG	2
	POP	ACC			;Not going to return, so pop return address.
	POP	ACC
	RJMP	STRSLT			;Just jump to results with interrupts disabled.
	
;--------Start here.-------------------------	

START:	OUTI	TCCR0B,0b00000001		;Set PWM clock to max.
	OUTI	TCCR0A,0b00000011		;Enable fast PWM, but on no ports.
	OUTI	PORTB,0
	;OUTI	DDRB,0x10
	;OUTI	PORTB,0x10		;Clear port B for OC output on OWIO.
	OUTI	PCMSK,0b00000100		;Interrupt on OWIO only.
	OUTI	GIMSK,0b00100000		;Enable interrupt on pin change.
	LDI	ZL,0x60			;Point Z to beginning of RAM.
	CLR	ZH			;(ZH should always be zero.)
	CLR	R0			;R0 = 0.

CLRLP:	ST	Z,R0			;Clear RAM & Reg locations, but not I/O space.
	INC	ZL			;Point to next. (No Z+ for no carry after 0xFF.)
	CPI	ZL,0x1E			;ZL points to ZL?
	BRNE	CLRLP			;  No:  Branch back.

;--------Protocol reset----------------------

RESET:	OUTI	SPL,RAMEND		;Stamp OW reset jumps here: Clear stack.

;--------Main program loop-------------------

MAINLP:	SER	RESULTH			;Initialize results to 0xFFFF.
	SER	RESULTL
	RCALL	GETBYTE			;Get a command from Stamp OWOUT.
	MOV	COMMAND,ACC
	ANDI	ACC,0xC0			;Is it a real command (both MSBs 0)?
	BRNE	TRYALT
	RJMP	DOCMD

TRYALT:	MOV	ACC,COMMAND
	ANDI	ACC,0xF0

TRYDMP:	SUBI	ACC,0xD0
	BRNE	TRYENT

DUMP:	MOV	ACC,COMMAND
	ANDI	ACC,0x0F
	JBRC	ACC,3,RDONE

	SUBI	ACC,0x0A
	BRNE	TRYID

	RCALL	GETBYTE
	CPI	ACC,0x20
	BRCC	RDZERO

	SUBI	ACC,-0x60
	MOV	ZL,ACC
	CLR	ZH

RDLP:	CPI	ZL,0x80
	BRCC	RDZERO
	LD	ACC,Z+
	RCALL	PUTBYTE
	RJMP	RDLP

RDZERO:	CLR	ACC
	RCALL	PUTBYTE
	RJMP	RDZERO

RDONE:	SUBI	ACC,-0x60
	MOV	ZL,ACC
	CLR	ZH
	LD	ACC,Z
	RCALL	PUTBYTE
	RJMP	MAINLP

TRYID:	SUBI	ACC,0x0D-0x0A
	BRNE	MAINLP

	LDI	ACC,DEVICE0
	RCALL	PUTBYTE
	LDI	ACC,DEVICE1
	RCALL	PUTBYTE
	LDI	ACC,VERSION
	RCALL	PUTBYTE
	RJMP	MAINLP

TRYENT:	SUBI	ACC,0xE0-0xD0
	BRNE	TRYFRQ

ENTER:	MOV	ACC,COMMAND
	ANDI	ACC,0x0F
	JBRC	ACC,3,WRONE

	SUBI	ACC,0x0A
	BRNE	MAINLP

	RCALL	GETBYTE
	CPI	ACC,0x20
	BRCC	WRNUL

	SUBI	ACC,-0x60
	MOV	ZL,ACC
	CLR	ZH

WRLP:	CPI	ZL,0x80
	BRCC	WRNUL
	RCALL	GETBYTE
	ST	Z+,ACC
	RJMP	WRLP

WRNUL:	RCALL	GETBYTE
	RJMP	MAINLP

WRONE:	SUBI	ACC,-0x60
	MOV	ZL,ACC
	CLR	ZH
	RCALL	GETBYTE
	ST	Z,ACC
	RJMP	MAINLP

TRYFRQ:	SUBI	ACC,0xF0-0xE0
	BREQ	SETFREQ
	RJMP	MAINLP

SETFREQ:	IN	ACC,TCCR0A
	ANDI	ACC,0xFD
	SBRS	COMMAND,0
	ORI	ACC,0x02
	OUT	TCCR0A,ACC
	ANDI	COMMAND,0x0F
	LSR	COMMAND
	INC	COMMAND
	OUT	TCCR0B,COMMAND
	RJMP	MAINLP

DOCMD:	LDI	PINMASK,0b11101111	;Set pin mask for pin 1.
	SBRC	COMMAND,PINNOH		;High bit of pin number set?
	LDI	PINMASK,0b11111010	;Yes: Set pin mask for pin 3.
	SBRS	COMMAND,PINNOL		;Low bit of pin number set?
	LSR	PINMASK			;  Yes: Shift mask right by 1.
	ORI	PINMASK,0b11100100	;Set all but candidate bits.

	JBRS	COMMAND,ANALOG,GETADC	;Is this an analog input command?

	JBRS	COMMAND,OUTPUT,SETOUT	;  No:  Is it an output command?

;--------Pin is a digital input--------------

GETINP:	IN	ACC,DDRB			;  No:  Make pin a digital input.
	AND	ACC,PINMASK
	OUT	DDRB,ACC
	IN	ACC,DIDR0		;Enable the digital input, too.
	AND	ACC,PINMASK
	OUT	DIDR0,ACC
	MOV	ACC,COMMAND		;Is it a counter input?
	ANDI	ACC,0x03
	BREQ	READPIN			;  No:  Simple digital input.

;--------Pin is a frequency input------------

READFRQ:	OUTI	GIFR,0xFF		;Clear the interrupt flags.
	SEI				;Set interupt for OWIO pin change (i.e. low->high)
	JBRS	COMMAND,PINNOH,COUNT32	;High bit of pin number set?

PINS10:	JBRS	COMMAND,PINNOL,COUNT10	;  No:  Low bit set?

COUNT00:	JBIC	PINB,3,COUNT00		;         No:  Count pin 0 (PB.3) until overflow or interrupt.
	SBIW	RESULTL,1
	BREQ	OVERFL
COUNT01:	JBIS	PINB,3,COUNT01
	RJMP	COUNT00

COUNT10:	JBIC	PINB,4,COUNT10		;         Yes: Count pin 1 (PB.4) until overflow or interrupt.
	SBIW	RESULTL,1
	BREQ	OVERFL
COUNT11:	JBIS	PINB,4,COUNT11
	RJMP	COUNT10

COUNT32:	JBRS	COMMAND,PINNOL,COUNT30	;  Yes: Low bit set?

COUNT20:	JBIC	PINB,1,COUNT20		;         No:  Count pin 2 (PB.1) until overflow or interrupt.
	SBIW	RESULTL,1
	BREQ	OVERFL
COUNT21:	JBIS	PINB,1,COUNT21
	RJMP	COUNT20

COUNT30:	JBIC	PINB,0,COUNT30		;         No:  Count pin 3 (PB.0) until overflow or interrupt.
	SBIW	RESULTL,1
	BREQ	OVERFL
COUNT31:	JBIS	PINB,0,COUNT31
	RJMP	COUNT30

OVERFL:	RJMP	OVERFL			;On overflow, don't count anymore. Just wait for interrupt.

STRSLT:	JBIC	PINB,OWIO,STRSLT
	COM	RESULTH			;Interrupt routine pops return and jumps here.
	COM	RESULTL			;Complement results.
	RJMP	COMPUTE			;Send appropriate bytes.

;--------Pin is a bit input.-----------------

READPIN:	COM	PINMASK			;Gets here with ACC = 0. Pin mask was mostly ones. Make a single one.
	IN	MACREG,PINB
	AND	MACREG,PINMASK		;Isolate the bit we're interested in.
	BREQ	SENDBIT			;Was it set?
	COM	ACC			;  Yes: Set ACC = 0xFF.
SENDBIT:	RCALL	PUTBIT			;Output the bit.
	RJMP	MAINLP			;Back for another command.

;--------Pin is an output.-------------------

SETOUT:	JBRC	COMMAND,PWM,SETHILO	;Is this a PWM output?

;--------Pin is PWM output.------------------
	
SETPWM:	RCALL	GETBYTE			;  Yes: Read the PWM value.
	COM	ACC
	JBRC	COMMAND,PINNOH,MAINLP	;Abort command if pin can't be a PWM out.

	JBRS	COMMAND,PINNOL,SETPWMA	;PWMA or PWMB?

SETPWMB:	OUT	OCR0B,ACC		;PWMB. Store value in counter.
	IN	ACC,TCCR0A		;Enable fast PWM on pin.
	ANDI	ACC,0xCF
	ORI	ACC,0x20
	SBRS	COMMAND,INVPWM
	ORI	ACC,0x10
	OUT	TCCR0A,ACC
	RJMP	SETDDR			;Make pin an output.	

SETPWMA:	OUT	OCR0A,ACC		;PWMA. Store value in counter.
	IN	ACC,TCCR0A		;Enable fast PWM on pin.
	ANDI	ACC,0x3F
	ORI	ACC,0x80
	SBRS	COMMAND,INVPWM
	ORI	ACC,0x40
	OUT	TCCR0A,ACC
	RJMP	SETDDR			;Make pin an output.

;--------Pin is a bit output.----------------

SETHILO:	JBRS	COMMAND,SETVAL,SETHIGH	;Setting pin high?

SETLOW:	IN	ACC,PORTB		;  No:  Set it low.
	AND	ACC,PINMASK
	RJMP	SETPIN

SETHIGH:	COM	PINMASK			;  Yes: Need a one in pin position.
	IN	ACC,PORTB
	OR	ACC,PINMASK		;       Set it high.
	COM	PINMASK			;       Restore mask.
	
SETPIN:	OUT	PORTB,ACC		;Output the pin setting.
	JBRC	COMMAND,PINNOH,SETDDR	;Is this one of the PWM-capable pins?

	IN	ACC,TCCR0A		;  Yes: Need to disable PWM.
	SBRC	COMMAND,PINNOL
	ANDI	ACC,0x3F
	SBRS	COMMAND,PINNOL
	ANDI	ACC,0xCF
	OUT	TCCR0A,ACC

SETDDR:	COM	PINMASK			;Make pin an output.
	IN	ACC,DDRB
	OR	ACC,PINMASK
	OUT	DDRB,ACC
	RJMP	MAINLP			;Back for another command.

;--------Pin is an analog input.-------------

GETADC:	IN	ACC,DDRB			;Make pin an input.
	AND	ACC,PINMASK
	OUT	DDRB,ACC
	COM	PINMASK			;Disable the digital input on this pin.
	IN	ACC,DIDR0		
	OR	ACC,PINMASK
	ORI	ACC,0x03
	OUT	DIDR0,ACC
	MOV	ACC,COMMAND		;Is this a comparator command?
	ANDI	ACC,0x03
	BREQ	GETCMP			;  Yes: Do comparator stuff.

;--------Pin is an ADC input.----------------

	JBRS	COMMAND,PINNOH,READY	;Can't read pins 2 or 3, so send 0xFFFF.
	
	OUTI	ADCSRA,0b10000110		;Enable the ADC with /64 clock.
	LDI	ACC,0b00000010		;Set ADMUX for right-justify on pin 1 (PB.4 = ADC2).
	SBRS	COMMAND,USELO		;Are we using ADCL?
	SETB	ACC,ADLAR		;  No:  Left justify result.
	SBRC	COMMAND,INTREF		;Using internal reference?
	SETB	ACC,REFS0		;  Yes: Set bit.
	SBRS	COMMAND,PINNOL		;Using pin 0?
	INC	ACC			;  Yes: Change mux to pin 0 (PB.3 = ADC3).
	OUT	ADMUX,ACC		;Set ADCMUX to new value.
	SBI	ADCSRA,ADSC		;Set ADSC bit to start conversion.
ADCWT:	RCALL	LOWTIME			;Check for resets during conversion.
	JBIS	ADCSRA,ADSC,ADCWT		;Branch back until conversion is complete.
	IN	RESULTL,ADCL		;Get conversion results.
	IN	RESULTH,ADCH
READY:	CLR	ACC			;Send the "not-busy" bit.
	RCALL	PUTBIT

COMPUTE:	JBRS	COMMAND,USEHI,SNDRSLT	;Using high byte of result?

	TST	RESULTH			;  No:  Is it non-zero?
	BREQ	SNDRSLT

	SER	RESULTL			;         Yes: Saturate low byte to 255.

SNDRSLT:	MOV	ACC,RESULTH		;Send results. Get high byte.
	SBRC	COMMAND,USEHI		;Using it?
	RCALL	PUTBYTE			;  Yes: Send it.
	MOV	ACC,RESULTL		;Get low byte.
	SBRC	COMMAND,USELO		;Using it?
	RCALL	PUTBYTE			;  Yes: Send it.
	RJMP	MAINLP			;Back for another command.

;--------Pin is comparator input.------------

GETCMP:	MOV	ACC,COMMAND		;Pin number 3?
	ANDI	ACC,0x30
	SUBI	ACC,0X30
	BREQ	PUTCMP			;  Yes: Can't use for compare. Send a zero.

	CBI	ADCSRA,ADEN		;Disable ADC.
	CLR	ACC			;Clear internal ref.
	SBRC	COMMAND,INTREF		;Comparing to internal ref?
	LDI	ACC,0x40			;  Yes: Set bit.
	OUT	ACSR,ACC			;Write to control register.
	JBRS	COMMAND,PINNOH,STDCMP	;Using standard comparator pins?

MUXCMP:	LDI	ACC,0b00000010		;  No:  Using an ADC pin.
	SBRS	COMMAND,PINNOL		;       Select the pin
	INC	ACC
	OUT	ADMUX,ACC		;       And write to ADMUX.
	SBI	ADCSRB,ACME		;       Tell comparator we're doing so.
	RJMP	DOCMP			;       Go do the comparison.

STDCMP:	CBI	ADCSRB,ACME		;  Yes: Tell the comparator.

DOCMP:	CLR	ACC			;Default zero.
	NOP				;Wait for synchronized input.
	SBIS	ACSR,ACO			;Is comparator bit set?
	SER	ACC			;  No:  Set ACC (reversed sense here).

PUTCMP:	RCALL	PUTBIT			;Send the result.		
	RJMP	MAINLP			;Back for another command.

;--------Subroutines.------------------------

GETBYTE:	LDI	BITCNT,8			;Get byte (ACC) from Stamp's OWOUT.
	RJMP	GETBITS

GETBIT:	LDI	BITCNT,1			;Get bit (ACC.7) from Stamp's OWOUT.

	
GETBITS:	RCALL	LOWTIME			;Wait for OWIO high or reset.
	CLR	ACC			;Clear value in case of just one bit.
GB_WT0:	JBIS	PINB,OWIO,GB_WT0		;Wait for OWIO to go low.		
	RCALL	LOWTIME			;Measure the pulse width or do reset.
	CPI	PTIME,64			;Set carry if pulse is narrower than 40uS (i.e. 1)
	ROR	ACC			;Rotate carry into ACC.
GB_ZERO:	DJNZ	BITCNT,GB_WT0		;Loop back for next bit, if any.
	RET				;All done.

;--------------------------------------

PUTBYTE:	LDI	BITCNT,8			;Send byte (ACC) to Stamp's OWIN.
	RJMP	PUTBITS

PUTBIT:	LDI	BITCNT,1			;Send bit (ACC.0) to Stamp's OWIN.


PUTBITS:	RCALL	LOWTIME			;Wait for OWIO high or reset.
PB_WT0:	JBIS	PINB,OWIO,PB_WT0		;Wait for OWIO to go low.
	JBRS	ACC,0,PB_NEXT		;Is next bit a 0?
	SBI	DDRB,OWIO		;  Yes: Assert OWIO output to pull low.
	LDI	PTIME,96			;       Set time for 30 uS.
PB_ZWT:	DJNZ	PTIME,PB_ZWT		;       Wait 96 * 3 clk / 9.6MHz = 30uS.
	CBI	DDRB,OWIO		;       Deassert OWIO output to float high.
PB_NEXT:	RCALL	LOWTIME			;Make sure OWIO actually goes high. May be reset.
	LSR	ACC			;Shift ACC right.
	DJNZ	BITCNT,PB_WT0		;Loop back for next bit, if any.
	RET				;Done.
	
;---------------------------------------	

LOWTIME:	CLR	PTIME			;Measure low width on OWIO in PTIME.
LT_TEST:	SBIC	PINB,OWIO		;Is pin low?		[1 clk]
	RET				;  No:  Return		[1 clk]
	NOP				;			[1 clk]
	IJNZ	PTIME,LT_TEST		;PTIME now > 255?          [3 clk]
LT_WT1:	JBIC	PINB,OWIO,LT_WT1		;  Yes: Low time exceeds 160uS, so reset.
	RJMP	RESET			;       Goto reset after OWIO goes high.
		
;+------------------------------------------------------------------------------------------------------------------------------+
;                                                   TERMS OF USE: MIT License                                                                                                              
;+------------------------------------------------------------------------------------------------------------------------------
;Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation     
;files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy,    
;modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software
;is furnished to do so, subject to the following conditions:                                                                   
;                                                                                                                              
;The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
;                                                                                                                              
;THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE          
;WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR         
;COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,   
;ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.                         
;+------------------------------------------------------------------------------------------------------------------------------+
			
